/* Copyright (c) 2015 Douglas Henke <dhenke@mythopoeic.org>
   This is free software -- see COPYING for details.           */
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <uWireM/uWireM.h>
#include <alpha15x4/alpha15x4.h>
#include <ds1307_rtc/ds1307_rtc.h>
#include "rotary_encoder.h"
#include "settings.h"

#ifdef NEVER
static void showhex(const uint8_t pos, const uint8_t val) {
   alpha15x4_draw(pos, val+(val < 10 ? '0' : 'A'-10), 0);
}

static void show(const char * const msg) {
   int i;
   for(i = 0; i < 4; i++) alpha15x4_draw(i, msg[i], 0);
}
#endif /* NEVER */

static void show4(const char c1, const char c2, const char c3, const char c4) {
   alpha15x4_draw(0, c1, 0);
   alpha15x4_draw(1, c2, 0);
   alpha15x4_draw(2, c3, 0);
   alpha15x4_draw(3, c4, 0);
}

/* Operational states for the user-interface state machine; see
   README.statemachine for details: */
typedef enum __attribute__((__packed__)) {
      /* All of the top-level menu choices need to be grouped together
         at the beginning of the enum, so they have consecutive,
         contiguous numeric values: */
   ST_TDISP=0, /* main time display */
   ST_SALRM,   /* set alarm "ALRM" prompt */
   ST_STIME,   /* set time "TIME" prompt */
   ST_DIMMR,   /* set brightness "DIM " prompt */
   /* end of top-level menu stuff */

   ST_ALRMH,   /* set alarm hours */
   ST_ALRMM,   /* set alarm minutes */
   ST_TIMEH,   /* set time hours */
   ST_TIMEM,   /* set time minutes */
   ST_DIMLV    /* set brightness level */
} state_t;

#define LAST_MENU_ITEM ST_DIMMR

int __attribute__((OS_main)) main(void) {
   button_evt_t btn = BTN_NONE;
   int8_t rotn, h, m;
   state_t state = ST_TDISP;
   settings_t settings;

   power_adc_disable();     /* turn off unused ADC to save power */

   renc_setup();               /* set up rotary encoder / push button */
   sei();                      /* enable global interrupts */

   TCCR1 |= _BV(CTC1);   /* put timer1 in compare then clear mode */
   TIMSK |= _BV(OCIE1A); /* send interrupt on timer1 match */
   
   uWireM_init();              /* set up I2C communication library */
   /* (no initialization needed for talking to DS1307 RTC) */

   settings_get(&settings);

   alpha15x4_init(settings.dim); /* set up 8x8 LED matrix */

   while(1) { /* main event loop -- runs forever */
      rotn = 0; btn = BTN_NONE; /* start with no events pending */

      /* We need to look at the event structure, decide what events
         we're going to handle, remember those events, and remove
         them from the event structure. To ensure that none of the
         interrupt handlers are reading or changing the event structure
         while we do this, we'll put the code in an atomic block (that
         runs with global interrupts disabled). We try to keep the code
         in the atomic block as fast as possible, since keeping interrupts
         disabled for too long risks missing rotary encoder steps... */
      ATOMIC_BLOCK(ATOMIC_FORCEON) {
         rotn = (_evt.rotary+1) >> 2;
         _evt.rotary -= (rotn << 2);
         if(_evt.btn != BTN_NONE) {
            btn = _evt.btn;
            _evt.btn = BTN_NONE;
         }
      }

      /* handle pending events (if any) and act according to current state */

      /* Long-press cancels out of any state except ST_TDISP. Handle
         this up front so we don't have to repeat it in every case: */
      if(state != ST_TDISP && btn == BTN_LONG) {
         state = ST_TDISP; btn = BTN_NONE; rotn = 0;
      }

      /* If we have rotation but no button presses and the current
         state is one of the top-level menu items, move along the menu: */
      if(btn == BTN_NONE && rotn && state <= LAST_MENU_ITEM) {
         rotn += state;
         while(rotn < 0) rotn += LAST_MENU_ITEM+1;
         while(rotn > LAST_MENU_ITEM) rotn -= LAST_MENU_ITEM+1;
         state = rotn; rotn = 0; 
      }
    
      switch(state) {
         case ST_TDISP: /* main time display */
            /* show the time as 24-hour HHMM */
            ds1307_get_hm(&h, &m);
            if(h < 0) { show4('-', '-', '-', '-'); break; }
            show4('0'+h/10, '0'+h%10, '0'+m/10, '0'+m%10);
            break;

         case ST_SALRM: /* set alarm prompt */
            if(btn == BTN_UP) {
               h = settings.alarmh; m = settings.alarmm;
               state = ST_ALRMH; break;
            }
            show4('A','L','R','M');
            break;

         case ST_STIME: /* set time prompt */
            if(btn == BTN_UP) {
               ds1307_get_hm(&h, &m);
               state = ST_TIMEH; break;
            }
            show4('T','I','M','E');
            break;

         case ST_DIMMR: /* set brightness prompt */
            if(btn == BTN_UP) {
               h = settings.dim;
               state = ST_DIMLV; break;
            }
            show4('D','I','M',' ');
            break;

         case ST_ALRMH: /* set alarm hours adjustment -- fall-through */
         case ST_TIMEH: /* set time hours adjustment */
            if(btn == BTN_UP) {
               state = (state == ST_TIMEH) ? ST_TIMEM : ST_ALRMM;
               break;
            }
            h += rotn;
            while(h > 23) h -= 24;
            while(h < 0) h += 24;
            show4('0'+h/10, '0'+h%10, ' ', ' ');
            break;

         case ST_ALRMM: /* set alarm minutes adjustment -- fall-through */
         case ST_TIMEM: /* set time minutes adjustment */
            if(btn == BTN_UP) {
               if(state == ST_TIMEM) {
                  ds1307_set_time(h, m, 0);
               } else {
                  settings.alarmh = h; settings.alarmm = m;
                  settings.alarm_enable = 1;
                  settings_store(&settings);
               }
               state = ST_TDISP; break;
            }
            m += rotn;
            while(m > 59) m -= 60;
            while(m < 0) m += 60;
            show4(' ', ' ', '0'+m/10, '0'+m%10);
            break;

         case ST_DIMLV: /* set brightness level adjustment */
            if(btn == BTN_UP) {
               settings.dim = h;
               settings_store(&settings);
               state = ST_TDISP; break;
            }
            h += rotn;
            while(h > 15) h -= 16;
            while(h < 0) h += 16;
            alpha15x4_dim(h);
            show4('*', '*', '0'+h/10, '0'+h%10);
            break;

      } /* end switch(state) */
   } /* end while(1) */
}
