/* Copyright (c) 2015 Douglas Henke <dhenke@mythopoeic.org>
   This is free software -- see COPYING for details.           */
#include <uWireM/uWireM.h>
#include <string.h>
#include <stdlib.h>
#include "ds1307_rtc.h"

#define DS1307_I2C_ADDR 0x68 /* I2C address */

/* register map */
#define DS_REG_SECONDS      0x00
#define DS_REG_MINUTES      0x01
#define DS_REG_HOURS        0x02
#define DS_REG_DAY_OF_WEEK  0x03
#define DS_REG_DAY_OF_MONTH 0x04
#define DS_REG_MONTH        0x05
#define DS_REG_YEAR         0x06
#define DS_REG_CONTROL      0x07

/* bits in DS_REG_CONTROL */
#define DS_CTRL_OUT  7
#define DS_CTRL_SQWE 4
#define DS_CTRL_RS1  1
#define DS_CTRL_RS0  0

/* bit 7 of DS_REG_SECONDS halts the clock if set to 1 */
#define DS_CLOCK_HALT 7

/* location and length of battery-backed RAM */
#define DS_RAM_START        0x08
#define DS_RAM_LEN          56

/*** ds1307_set_time() -- set local time of day to a specified value

Adjusts the time-of-day portion of the clock to a specified value, and
starts the clock if it is not already running. Resets the divider chain
(so the specified second will start within a few dozen microseconds of
when this function is called).

Arguments:
   h -- hours part of local time (in 24-hour format), from 0 to 23 inclusive
   m -- minutes part of local time, from 0 to 59 inclusive
   s -- seconds part of local time, from 0 to 59 inclusive

Note: assumes 24-hour time.
***/
void ds1307_set_time(const uint8_t h, const uint8_t m, const uint8_t s) {
   uint8_t buf[5];

   /* FIXME: the division and modulo code emitted by GCC isn't great.
      This function could probably benefit from one of the approaches
      in "AVR200: Multiply and Divide Routines":
         http://www.atmel.com/Images/doc0936.pdf

      I already tried a naive C implementation of divide by ten a la:
         q = (((uint16_t)n * (uint16_t)0xcd) >> 8) >>3;
      but this actually yielded larger code. */
 
   buf[0] = uWireM_addr_send(DS1307_I2C_ADDR);
   buf[1] = DS_REG_SECONDS; /* address where write will start */
   buf[2] = (s/10) << 4 | (s%10); /* DS_REG_SECONDS */
   buf[3] = (m/10) << 4 | (m%10); /* DS_REG_MINUTES */
   buf[4] = h%10;                 /* DS_REG_HOURS */
   if(h >= 20) buf[4] |= (1 << 5);
   else if(h >= 10) buf[4] |= (1 << 4);

   uWireM_xfer(buf, 5);
}

/*** ds1307_get_hm() -- get hours and minutes parts of current time

Retrieves the hours and minutes parts of the local time.

Arguments:
   h -- pointer to int8_t into which hours value (0-24) will be stored
   m -- pointer to int8_t into which minutes value (0-59) will be stored

Note: If the clock is halted, the value pointed to by h will be set to
an out-of-band value (-1) and the value pointed to by m will be left
unmodified.
***/
void ds1307_get_hm(int8_t * const h, int8_t * const m) {
   uint8_t buf[4];

   /* Set address register (the place we'll start reading) to DS_REG_SECONDS: */
   buf[0] = uWireM_addr_send(DS1307_I2C_ADDR);
   buf[1] = DS_REG_SECONDS;
   uWireM_xfer(buf, 2);

   /* Read three registers, starting with DS_REG_SECONDS: */
   buf[0] = uWireM_addr_recv(DS1307_I2C_ADDR);
   uWireM_xfer(buf, 4);

   /* If the clock is halted, assume it is unset since the last total loss
      of power. Set hours to -1 and return. */
   if(buf[1] & _BV(DS_CLOCK_HALT)) { *h=-1; return; }

   /* Pull hours and minutes out of register fields: */
   *h = buf[3] & 0x0f;
   if(buf[3] & (1<<5)) *h += 20;
   else if(buf[3] & (1<<4)) *h += 10;
   *m = (buf[2] >> 4)*10 + (buf[2] & 0x0f);
}

/*** ds1307_ram_store() -- store a structure in the DS1307 battery-backed RAM

Stores data in the DS1307 battery-backed RAM.

Arguments:
   p -- pointer to data to be stored, which must be a composite type
      the *very first* fields of which must be DS1307_RAM_XFER_FLDS. Note
      that the contents of these initial fields will be modified.
   len -- length of the data to be stored, in bytes, including the
      DS1307_RAM_XFER_FLDS; must be greater than 2 and less than or
      equal to DS_RAM_LEN+2
***/
void ds1307_ram_store(void * const p, const int8_t len) {
   uint8_t *buf = (uint8_t *)p;

   buf[0] = uWireM_addr_send(DS1307_I2C_ADDR);
   buf[1] = DS_RAM_START;
   uWireM_xfer(buf, len);
}

/*** ds1307_ram_fetch() -- retrieve structure from DS1307 battery-backed RAM

Retrieves data previously stored in the DS1307 battery-backed RAM using
ds_ram_store() (above).

Arguments:
   p -- pointer to data structure to be used to hold the retrieved data.
      This must be a composite type the *very first* fields of which must
      be DS1307_RAM_XFER_FLDS. (Note that the contents of these initial
      fields may differ from what was in them when ds1307_ram_store() was
      called. These fields should be treated as opaque by the caller.)
   len -- length of the data to be retrieved, in bytes, including the
      DS1307_RAM_XFER_FLDS; must be greater than 2 and less than or
      equal to DS_RAM_LEN+2
***/
void ds1307_ram_fetch(void * const p, const int8_t len) {
   uint8_t *buf = (uint8_t *)p;

   /* Set address register (the place we'll start reading) to DS_RAM_START: */
   buf[0] = uWireM_addr_send(DS1307_I2C_ADDR);
   buf[1] = DS_RAM_START;
   uWireM_xfer(buf, 2);

   buf[1] = uWireM_addr_recv(DS1307_I2C_ADDR);
   uWireM_xfer(buf+1, len-1);
}
